/*
 *
 * Reads DCF77 Clock at parallel port ("PC-Funkuhr") via 
 * ioctl command LPPCFTIMER. Needs patch pcf.linux-<VERSION>.diff 
 * to the kernel.
 *
 * Matthias Schtze                      30 Aug 98, 23 Apr 99
 *
 */
#include "pcf77.h"

char portname[SSTRLEN] = STDPORT;
char progname[SSTRLEN];

char timeinfo[NUMBYTES] = {GETPCF77TIME,0,0,0,0,0,0};

char mailto[LSTRLEN + 1] = "";
char scalls[LSTRLEN + 1] = "";

/*
 *	setprob means: set time even if there were problems receiving radio signal
 */
int quiet = FALSE,timeset = FALSE,hardset = FALSE,setprob = FALSE;
int autoprobe = FALSE,endflag = FALSE,printdiff = FALSE;

int qualitycheck = FALSE;

/*
 * print out usage info
 */
void usage(void)
{
  fprintf(stderr,"usage: %s [options] [+FORMAT]\n",progname);
  fprintf(stderr,"   reads timer information of DCF77-Clock (\"PC-Funkuhr\") at printer port,\n");
  fprintf(stderr,"   needs an appropriate patched lp-kernel-module.\n");
  fprintf(stderr,"   See date(1) for a description of the FORMAT-string.\n");
  fprintf(stderr,"options:\n");
  fprintf(stderr,"   -a       : autoprobe lp devices\n");
  fprintf(stderr,"   -c       : check quality of DCF77 radio signal\n");
  fprintf(stderr,"   -D       : print out difference between system and DCF time\n");
  fprintf(stderr,"   -h, -?   : prints this usage-message\n");
  fprintf(stderr,"   -m <addr>: mail problems with the \"PC-Funkuhr\" to addr\n");
  fprintf(stderr,"   -p <dev> : Printer Port (default: %s)\n",portname);
  fprintf(stderr,"   -q       : quiet, don't print time read from clock\n");
  fprintf(stderr,"   -s       : set system clock \"smoothly\",\n");
  fprintf(stderr,"              if time-difference is less than %d seconds\n",MAXSMOOTHADJUST);
  fprintf(stderr,"   -S       : set system clock immediately\n");
  fprintf(stderr,"   -t       : same as -s, don't set time if \"PC-Funkuhr\"\n");
  fprintf(stderr,"              had problems receiving radio signal\n");
  fprintf(stderr,"   -T       : same as -S, don't set time if \"PC-Funkuhr\"\n");
  fprintf(stderr,"              had problems receiving radio signal\n");
  fprintf(stderr,"   -v       : print version and exit\n");
  fprintf(stderr,"\n");
  exit(1);
}

/*
 * subtract two timeval values
 */
int timeval_subtract (result, x, y)
struct timeval *result, *x, *y;
{
  /* Perform the carry for the later subtraction by updating Y. */
  if (x->tv_usec < y->tv_usec) {
	int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
	y->tv_usec -= 1000000 * nsec;
	y->tv_sec += nsec;
  }
  if (x->tv_usec - y->tv_usec > 1000000) {
	int nsec = (y->tv_usec - x->tv_usec) / 1000000;
	y->tv_usec += 1000000 * nsec;
	y->tv_sec -= nsec;
  }
  
  /* Compute the time remaining to wait.
	 `tv_usec' is certainly positive. */
  result->tv_sec = x->tv_sec - y->tv_sec;
  result->tv_usec = x->tv_usec - y->tv_usec;
  
  /* Return -1 if result is negative else 1. */
  return (x->tv_sec < y->tv_sec) ? -1 : 1;
}

/*
 * time between talking to DCF clock and the execution of some part of this 
 * code may lead to differences of at least one second
 * adjust this difference:
 *   bt is the pointer to the broken time read from DCF clock
 *   TP is the pointer to the calendar time at the time we read DCF clock
 *   msec are the milliseconds read from DCF clock
 *
 * This function produces wrong results if more than 1000 msecs
 * are elapsed at 23:59:59 at the last day of a month
 * please fix if you like
 */
int adjust_seconds(struct tm *bt,struct timeval *TP,int *msec)
{
  struct timeval TR;
  
  /*
   * testing this function...
   */
  /*	*msec = 0;
   *	return 0;
   */
  /*
   * end testing....
   */
  
  gettimeofday(&TR,NULL);

#ifdef STD_DBG
  fprintf(stderr,"sec=%ld -- usec=%ld\n",TR.tv_sec,TR.tv_usec);
#endif
  
  timeval_subtract(&TR,&TR,TP);
  *msec += TR.tv_usec / 1000;
  if(*msec > 1000)
  {
	bt -> tm_sec  += *msec / 1000;
	if(bt -> tm_sec > 59)
	{
	  bt -> tm_sec -= 60;
	  bt -> tm_min++;
	  if(bt -> tm_min > 59)
	  {
		bt -> tm_min -= 60;
		bt -> tm_hour++;
		if(bt -> tm_hour > 23)
		{
		  bt -> tm_hour -= 24;
		  bt -> tm_mday++;
		}
	  }
	}
	*msec %= 1000;
  }
  return 0;
}

void signal_handler(int sig)
{
  endflag = TRUE;
}

/*
 *	get signals and to be able to end gracefully
 */
void init_sighandler(void)
{
  if (signal(SIGTERM,signal_handler) == SIG_ERR )
    fprintf(stderr,"Could not install SIGTERM handler.\n");

  if (signal(SIGINT,signal_handler) == SIG_ERR )
    fprintf(stderr,"Could not install SIGINT handler.\n");

  if (signal(SIGHUP,signal_handler) == SIG_ERR )
    fprintf(stderr,"Could not install SIGHUP handler.\n");
}


/*
 *	Quality check is done by controlling the original DCF77 signal which
 *	is copied to timeinfo[0] by the kernel
 *
 *	Since I really don't know much about the radio signal this implementation
 *	is rather simple: The two sorts of bits are distinguished by their
 *	length: a 0-Bit has a length of 100 ms, a 1-Bit a length of 200 ms
 *	The idea is to check signal lengths (duration of the "1" in timeinfo[0])
 *	and then to look whether these durations are 100 or 200 ms. If this is
 *	true, we believe that we receive the radio signal.
 *
 *	The output here is quite simple: Valid values result in a "#" Output
 *	invalid ones in "-".
 */
void quality_out(int handle)
{
#define OUTLEN 40
#define CHECKBASIS 20
  int i,j = CHECKBASIS / 2,cnt = 0,valid[CHECKBASIS],cannotbevalid;
  struct timeval TP,TQ,TR;  

  /*
   *	Preparation of the line
   *	0%]------------------------------------[100%
   *	we start with a 50% signal quality
   */
  printf("press Ctrl-C to end quality check!\n");
  printf("  ");
  for(i = 0;i < CHECKBASIS;i++)
	valid[i] = i & 1; /* 50% as preselection */
  for(i = 0;i < OUTLEN;i++)
	printf("%c",(i % (OUTLEN / 4)) ? ' ' : '|');

  printf(" |\n0%%]");
  for(i = 0;i < OUTLEN;i++)
	printf("-");
  printf("[100%%");
  for(i = 0;i < OUTLEN + 5;i++)
	printf("\b");
  fflush(stdout);

  /*
   *	synchronize to the DCF77 signal:
   */
  if(timeinfo[0])
	while(timeinfo[0])
	{
	  timeinfo[0] = CONTQUALCHEK;
	  ioctl(handle, LPPCFTIMER, timeinfo);
	  usleep(DCF_SIGNAL_WAIT);
	}

  /*
   *	and now continuously check quality
   */
  for(cnt = 0;!endflag;cnt++)
  {
	cannotbevalid = FALSE;
	gettimeofday(&TP,NULL);
	while (!timeinfo[0])
	{
	  timeinfo[0] = CONTQUALCHEK;
	  ioctl(handle, LPPCFTIMER, timeinfo);
	  usleep(DCF_SIGNAL_WAIT);
	}
	gettimeofday(&TQ,NULL);
	timeval_subtract(&TR,&TQ,&TP);
	/* too much time between two signals (either Bit 0 or Bit 1)		*/
	if(TR.tv_sec)
	  cannotbevalid = TRUE;

	/* check bitlength	*/
	gettimeofday(&TP,NULL);
	while( timeinfo[0] )
	{
	  timeinfo[0] = CONTQUALCHEK;
	  ioctl(handle, LPPCFTIMER, timeinfo);
	  usleep(DCF_SIGNAL_WAIT);
	}
	gettimeofday(&TQ,NULL);
	timeval_subtract(&TR,&TQ,&TP);
	if(((TR.tv_usec < DCF_BITLEN1 + 2 * DCF_SIGNAL_WAIT) &&
		(TR.tv_usec > DCF_BITLEN1 - 2 * DCF_SIGNAL_WAIT)) ||
	   ((TR.tv_usec < DCF_BITLEN0 + 2 * DCF_SIGNAL_WAIT) &&
		(TR.tv_usec > DCF_BITLEN0 - 2 * DCF_SIGNAL_WAIT)))
	  valid[cnt % CHECKBASIS] = 1; else valid[cnt % CHECKBASIS] = 0;
	if(cannotbevalid)
	  valid[cnt % CHECKBASIS] = 0;
	/*
	 *	finally print out result
	 */
	for(i = 0,j = 0;i < CHECKBASIS;i++)
	  if(valid[i]) j++;
	for(i = 0;i < (j * OUTLEN) / CHECKBASIS;i++)
	  printf("#");
	for(;i < OUTLEN;i++)
	  printf("-");
	fflush(stdout);
	for(i = 0;i < OUTLEN;i++)
	  printf("\b");
  }
  j = (j * 100) / CHECKBASIS;
  printf("\n");
  printf("Quality check ended with a %d%% signal strength,\n",j);
  printf("which %s to get a secure DCF time info!\n\n",
		 (j < 50) ? "is too small" :
		 ((j < 80) ? "may be sufficient" : "is ok"));
  /*
   *	The DCF clock seems to require such a "reset":
   */
  for(i = 0;i < TRIALS;i++)
  {
	timeinfo[0] = GETPCF77TIME;
	ioctl(handle, LPPCFTIMER, timeinfo);
  }
}

/*
 * read out DCF clock at printer port portname
 *
 * The values in timeinfo are (BCD coding of time and date):
 * timeinfo[0]: Bit 0:   1 = radio signal received,0 = did not receive signal
 *              Bit 1-3: 000
 *              Bit 4-7: 1001
 * timeinfo[1]: Bit 0-3: seconds, high nibble
 *              Bit 4-7: seconds, low nibble
 * timeinfo[2]: Bit 0-3: minutes, high nibble
 *              Bit 4-7: minutes, low nibble
 * timeinfo[3]: Bit 0-3: hours, high nibble
 *              Bit 4-7: hours, low nibble
 * timeinfo[4]: Bit 0-3: Day of week
 *              Bit 4:   1=DST
 *              Bit 5:   1=no DST
 *              Bit 6:   low Battery
 *              Bit 7:   unused
 * timeinfo[5]: Bit 0-3: day, high nibble
 *              Bit 4-7: day, low nibble
 * timeinfo[6]: Bit 0-3: month, high nibble
 *              Bit 4-7: month, low nibble
 * timeinfo[7]: Bit 0-3: year, high nibble
 *              Bit 4-7: year, low nibble
 * timeinfo[8]: Bit 3:   0 = lower half of second, 1 = upper half of second
 *              Bit 4-7: 1/32 of a second
 *
 * This data is taken from the documentation of the "PC-Funkuhr". In contrast
 * to this documentation, the bit-order of each byte is changed from left to 
 * right (just as data has been read from the "PC-Funkuhr").
 *
 * The parameter witherrors denotes whether or not to display error messages
 * when trying to access port portname.
 */
int talktoclock(char *portname,int witherrors)
{
#define pbSIZE 512
  char printbuf[pbSIZE];
  int handle,trial,retval = -1;
  struct stat statbuf;	
  int msec = 0;
  struct timeval TP,TQ,TR;
  struct tm      bt;

  /* open port																*/
  if((handle = open(portname, O_RDWR, 0)) < 0)
  {
	if(witherrors)
	  perror(portname);
    return -1;
  }
  
  /* check, whether this is a valid port type								*/
  fstat(handle, &statbuf);
  if((!S_ISCHR(statbuf.st_mode)) || (MAJOR(statbuf.st_rdev) != 6 )
     || (MINOR(statbuf.st_rdev) > 3)) {
    fprintf(stderr,"%s: %s not an lp device.\n", progname, portname);
	return -1;
  }

  if(qualitycheck)
	timeinfo[0] = CHECKQUALITY;
  
  /* because of some "deaf" milliseconds of the clock, we try three times
     also check time of day at the moment we talk to the clock in order
	 to compute difference between time reading and printing 				*/
  for(trial = 0;trial < TRIALS;trial++)
  {
	gettimeofday(&TP,NULL);
	retval = ioctl(handle, LPPCFTIMER, timeinfo);
	if(!retval) break;
  }
  /* well, there seems to be no "PC-Funkuhr" at portname					*/
  if (retval < 0)
  {
	if(witherrors)
	  perror("LPPCFTIMER error");
	close(handle);
	return(retval);
  }


  if(qualitycheck)
  {
	quality_out(handle);
	close(handle);
	return 0;
  }
  
  close(handle);

  /* compute date															*/
  bt.tm_mday  = (timeinfo[5] & 0x0f) * 10 + ((timeinfo[5] & 0xf0) >> 4);
  bt.tm_mon   = (timeinfo[6] & 0x0f) * 10 + ((timeinfo[6] & 0xf0) >> 4) - 1;
  bt.tm_year  = (timeinfo[7] & 0x0f) * 10 + ((timeinfo[7] & 0xf0) >> 4);
  /*
   * add 100 to the two-digit year transmitted from PC-Funkuhr if it's
   * lower than CENTURYYEAR 
   */
  bt.tm_year += (bt.tm_year < CENTURYYEAR) ? 100 : 0;
  bt.tm_wday  = timeinfo[4] & 0x0f;

  bt.tm_sec   = (timeinfo[1] & 0x0f) * 10 + ((timeinfo[1] & 0xf0) >> 4);
  msec = ((timeinfo[8] & 0x08) ? 500 : 0) + ((timeinfo[8] & 0xf0) >> 4) \
	* 1000 / 32;
  bt.tm_min   = (timeinfo[2] & 0x0f) * 10 + ((timeinfo[2] & 0xf0) >> 4);
  bt.tm_hour  = (timeinfo[3] & 0x0f) * 10 + ((timeinfo[3] & 0xf0) >> 4);
  bt.tm_isdst = (timeinfo[4] & 0x10) ? 1 :  ((timeinfo[4] & 0x20) ? 0 : -1);

#ifdef STD_DBG
  sleep(1);
#endif
  
  /*
   * calculate difference between system and DCF time
   */
  adjust_seconds(&bt,&TP,&msec);
  TQ.tv_sec  = mktime(&bt);
  TQ.tv_usec = msec * 1000;
  gettimeofday(&TR,NULL);
  retval = timeval_subtract(&TR,&TQ,&TR);

  /*
   * setting new (DCF-) time
   * if there is a moderate difference between actual time of computer and
   * DCF-time, slowly adjust time using adjtime, otherwise set it using
   * settimeofday
   *
   * However, if the clock had trouble receiving radio signal and one of the
   * -t or -T option is set, then do not set time
   */
  if(timeset && !((timeinfo[0] & 0x01) && !setprob))
  {
	if(hardset)
	{
	  if(settimeofday(&TQ,NULL) < 0)
		perror("settimeofday");
	}
	else if (TR.tv_sec < MAXSMOOTHADJUST && TR.tv_sec > -MAXSMOOTHADJUST)
	{
	  if(adjtime(&TR,NULL) < 0)
		perror("adjtime");
	}
	else
	{
	  sendwarning("Too large time difference for adjusting system clock!");
	}
	if(!quiet) printf("system time has been set to DCF time\n");
  }

  /*
   * time set done, then print out what we have read from the PC-Funkuhr
   */
  if(!quiet)
  {
	adjust_seconds(&bt,&TP,&msec);
	strftime(printbuf,pbSIZE,pbTEMPL,&bt);
	printf("%s",printbuf);
	  
	if(printdiff)
	  printf(" (diff=%+6ldus)",-TR.tv_sec * 1000000 - TR.tv_usec);
	printf("\n");
  }

  /*
   * check information given by the "PC-Funkuhr" and issue warnings
   * either via mail or to stderr
   */
  if (timeinfo[4] & 0x40)
	sendwarning("Battery of \"PC-Funkuhr\" low! Get new batteries!");
  if (timeinfo[0] & 0x01)
	sendwarning("Did not receive DCF radio signal last time tried!");
  
  return 0;
}

/*
 * send out a warning, either to stderr or via mail
 * do nothing when "warning" is empty
 */
int sendwarning(char *warning)
{
  if(!strlen(warning)) return 1;
  
  if(strlen(mailto))
	{
	  snprintf(scalls,LSTRLEN,"echo %s warning: %s | mail -s %s %s",progname,
			  warning,progname,mailto);
	  system(scalls);
	}
  else
	fprintf(stderr,"%s warning: %s\n",progname,warning);
  return 0;
}


/*
 * check all lp devices /dev/lp? whether the DCF77 clock is connected
 */
int talktoalldevices(void)
{
  char lp[50];
  int i;
  DIR *dp,*d1;
  struct dirent *ep;
     
  if((d1 = dp = opendir ("/dev")) != NULL)
  {
	for(i = 0;i < MAXLPPORTS;i++)
	{
	  dp = d1;
	  sprintf(lp,"lp%d",i);
	  while ((ep = readdir (dp)) != NULL)
		if(!strcmp(ep->d_name,lp))
		{
		  sprintf(lp,"/dev/lp%d",i);
		  printf("%s: check for port %s\n",progname,lp);
		  if(!talktoclock(lp,FALSE))
		  {
			printf("\"PC-Funkuhr\" found at port %s\n",lp);
			(void) closedir (dp);
			return 0;
		  }
		  else
			printf("\n");
		  break;
		}
	}
	(void) closedir (dp);
  }
  else
  {
	perror("/dev");
	return -1;
  }
  return 0;
}

int main(int argc, char **argv)
{
  int c,ret = 0;
  extern int opterr,optind;
  extern char *optarg;

  strcpy(progname,BASENAME(argv[0]));
  while ((c = getopt(argc, argv, "acDm:p:qsStTvh?")) != EOF)
    switch (c) 
    {
	case 'a':
	  autoprobe = TRUE;
	  break;
	case 'c':
	  qualitycheck = TRUE;
	  break;
	case 'D':
	  printdiff = TRUE;
	  break;
	case 'm':
	  strncpy(mailto,optarg,LSTRLEN);
	  break;
	case 'S':
	case 's':
	case 't':
	case 'T':
	  if(getuid())
		fprintf(stderr,"%s: must be root to set time!\n", progname);
	  else
		timeset = TRUE;
	  hardset = setprob = FALSE;
	  if(c == 'S' || c == 'T')
		hardset = TRUE;
	  if(c == 's' || c == 'S')
		setprob = TRUE;
	  break;
	case 'p':
	  strncpy(portname,optarg,SSTRLEN);
	  break;
	case 'q':
	  quiet = TRUE;
	  break;
	case 'v':
	  printf("%s: Version %d.%d.%d\n",progname,VERSION,SUB_VERSION,REVISION);
	  exit(0);
	  break;
	case 'h':
	case '?':
	default:
	  usage();
	}

  for (c = optind; c < argc;c++)
	if(argv[c][0] == '+')
	  strncpy(pbTEMPL,&argv[c][1],LSTRLEN);
	else
	  usage();

  init_sighandler(); /* use our own signal handler */

  if(autoprobe)
	ret = talktoalldevices();
  else
	ret = talktoclock(portname,TRUE);
  return ABS(ret);
}












